from typing import *

from wizwalker import XYZ, Keycode
from wizwalker.client import Client
from wizwalker.memory import DynamicClientObject


class SprintyClient(Client):
    async def teleport(self, xyz: XYZ, yaw: int = None, turn_time: float = 0.1):
        """
        Teleports the client and moves slightly to update the position
        :param xyz: The target position
        :param yaw: Yaw at target
        :param turn_time: Time spent turning left to update the position
        """
        await super().teleport(xyz, yaw)
        if turn_time > 0:
            await self.send_key(Keycode.A, turn_time)

    async def get_base_entities_with_vague_name(self, name: str) -> List[DynamicClientObject]:
        async def _pred(e):
            if temp := await e.object_template():
                return name.lower() in (await temp.object_name()).lower()
            return False

        return await self.get_base_entities_with_predicate(_pred)

    async def get_base_entities_with_behaviors(self, behaviors: List[str]):
        res = []
        for entity in await self.get_base_entity_list():
            good = True
            inactive_behaviors = [await beh.read_type_name() for beh in await entity.inactive_behaviors()]
            for behavior in behaviors:
                if behavior not in inactive_behaviors:
                    good = False
                    break
            if good:
                res.append(entity)
        return res

    async def get_health_wisps(self) -> List[DynamicClientObject]:
        return await self.get_base_entities_with_vague_name("WispHealth")

    async def get_mana_wisps(self) -> List[DynamicClientObject]:
        return await self.get_base_entities_with_vague_name("WispMana")

    async def get_mobs(self) -> List[DynamicClientObject]:
        async def _pred(e: DynamicClientObject):
            behaviors = await e.inactive_behaviors()
            for b in behaviors:
                if (await b.read_type_name()) == "NPCBehavior":
                    return await b.read_value_from_offset(288, "bool")
            return False
        return await self.get_base_entities_with_predicate(_pred)

    async def find_closest_of_entities(self, entities: List[DynamicClientObject]) -> Optional[DynamicClientObject]:
        closest = None
        smallest_dist = 0
        self_pos = await self.body.position()
        for w in entities:
            dist = self_pos.distance(await w.location())
            if closest is None or dist < smallest_dist:
                smallest_dist = dist
                closest = w
        return closest

    async def find_closest_by_predicate(self, pred: Callable) -> Optional[DynamicClientObject]:
        return await self.find_closest_of_entities(await self.get_base_entities_with_predicate(pred))

    async def find_closest_by_name(self, name: str) -> Optional[DynamicClientObject]:
        return await self.find_closest_of_entities(await self.get_base_entities_with_name(name))

    async def find_closest_by_vague_name(self, name: str) -> Optional[DynamicClientObject]:
        return await self.find_closest_of_entities(await self.get_base_entities_with_vague_name(name))

    async def find_closest_health_wisp(self) -> Optional[DynamicClientObject]:
        return await self.find_closest_of_entities(await self.get_health_wisps())

    async def find_closest_mana_wisp(self) -> Optional[DynamicClientObject]:
        return await self.find_closest_of_entities(await self.get_mana_wisps())

    async def find_closest_mob(self) -> Optional[DynamicClientObject]:
        return await self.find_closest_of_entities(await self.get_mobs())

    async def tp_to_closest_of(self, entities: List[DynamicClientObject]):
        if e := await self.find_closest_of_entities(entities):
            await self.teleport(await e.location())
            return True
        return False

    async def tp_to_closest_by_name(self, name: str) -> bool:
        if e := await self.find_closest_by_name(name):
            await self.teleport(await e.location())
            return True
        return False

    async def tp_to_closest_by_vague_name(self, name: str) -> bool:
        if e := await self.find_closest_by_vague_name(name):
            await self.teleport(await e.location())
            return True
        return False

    async def tp_to_closest_health_wisp(self) -> bool:
        return await self.tp_to_closest_of(await self.get_health_wisps())

    async def tp_to_closest_mana_wisp(self) -> bool:
        return await self.tp_to_closest_of(await self.get_mana_wisps())

    async def tp_to_closest_mob(self) -> bool:
        return await self.tp_to_closest_of(await self.get_mobs())

    async def calc_health_ratio(self) -> float:
        """Simply returns current health divided by max health"""
        return await self.stats.current_hitpoints() / await self.stats.max_hitpoints()

    async def calc_mana_ratio(self) -> float:
        """Simply returns current health divided by max health"""
        return await self.stats.current_mana() / await self.stats.max_mana()

    async def has_potion(self) -> bool:
        return await self.stats.potion_charge() > 1.0

    async def use_potion(self) -> bool:
        if await self.has_potion():
            await self.send_hotkey([Keycode.CTRL], Keycode.O)
            return True
        return False

    async def needs_potion(self, health_percent: int = 20, mana_percent: int = 10) -> bool:
        return await self.calc_health_ratio() * 100 <= health_percent or \
               await self.calc_mana_ratio() * 100 < mana_percent

    async def use_potion_if_needed(self, health_percent: int = 20, mana_percent: int = 10) -> bool:
        """
        :param health_percent: How low health needs to be in percent
        :param mana_percent: How low mana needs to be in percent
        :return: False if it needed a potion but doesn't have any and True if it needed no potion or used one
        """
        if await self.needs_potion():
            return await self.use_potion()
        return True
