from typing import *
from dataclasses import dataclass
from enum import Enum, auto

import wizwalker
from wizwalker.combat import CombatHandler
from wizwalker.combat import CombatMember
from wizwalker.combat.card import CombatCard

from .Exceptions import TracebacklessException


class TargetType(Enum):
    Self = auto()
    Boss = auto()
    Enemy = auto()
    Ally = auto()
    AOE = auto()


@dataclass  # maybe freeze?
class MoveConfig:
    move: str
    enchant: str
    target: Tuple[TargetType, int]


class CombatConfigProvider:
    class ParsingError(TracebacklessException):
        """Raised when the config parser encounters something it doesn't understand"""
        pass

    def __init__(self, path: str):
        self.rounds = []
        self.filename = path
        with open(self.filename) as file:
            self.data = self.parse_config(file.read())

    # TODO: Make/Use real parser
    def parse_config(self, data: str) -> List[List[MoveConfig]]:  # parse full config
        contents = data.splitlines()
        res = []
        line_num = 0  # for error reporting
        for line in contents:
            line_num += 1

            line = line.strip()
            if len(line) < 1:
                continue

            res.append(self.parse_line(line, line_num))
        return res

    def parse_line(self, line: str, line_num: int) -> List[MoveConfig]:
        """Parses a single config line, may be useless for other formats"""
        res = []

        priorities = [priority.strip() for priority in line.split(";")]  # parse slackaduts' config format (extended)
        for priority in priorities:
            parts = [p.strip() for p in priority.split(",")]
            if len(parts) > 3:
                raise self.ParsingError(
                    f"Might be missing semicolon in config \"{self.filename}\" line {line_num}:\n\t{line}")

            try:
                move, enchant, target = parts
            except Exception:
                raise self.ParsingError(
                    f"Might be missing comma in config \"{self.filename}\" line {line_num}:\n\t{line}")

            res.append(MoveConfig(move=move, enchant=enchant, target=self.parse_target(target, line, line_num)))

        return res

    def parse_target(self, target: str, full_line: str, line_num: int) -> Tuple[TargetType, int]:
        target_type: TargetType
        target_extra = None

        specifier = target.split("(")[0]

        try:
            # "simple" ones
            if specifier == "self":
                target_type = TargetType.Self
            elif specifier == "boss":
                target_type = TargetType.Boss
            elif specifier == "aoe":
                target_type = TargetType.AOE

            # special ones
            elif specifier == "enemy":
                target_type = TargetType.Enemy
                if "(" in target:
                    if ")" not in target:
                        raise self.ParsingError(
                            f"Might be missing closing parenthesis in config "
                            f"\"{self.filename}\" line {line_num}:\n\t{full_line}\n\t\t-> {target}")
                    target_extra = int(target.split("(")[1][:-1])
            elif specifier == "ally":
                target_type = TargetType.Ally
                if "(" in target:
                    if ")" not in target:
                        raise self.ParsingError(
                            f"Might be missing closing parenthesis in config "
                            f"\"{self.filename}\" line {line_num}:\n\t{full_line}\n\t\t-> {target}")
                    target_extra = int(target.split("(")[1][:-1])

            else:  # bad target
                raise self.ParsingError(
                    f"Non-existant target specifier in config "
                    f"\"{self.filename}\" line {line_num}:\n\t{full_line}\n\t\t-> {target}")
        except Exception as e:
            if e is not self.ParsingError:  # this is something that was not anticipated
                raise self.ParsingError(
                    f"Bad target specified in config "
                    f"\"{self.filename}\" line {line_num}:\n\t{full_line}\n\t\t-> {target}")
            raise

        return target_type, target_extra

    def get_round_or_none(self, r: int) -> List[MoveConfig]:
        if len(self.data) < r:
            return []
        return self.data[r]

    def defined_rounds(self) -> int:
        return len(self.data)


class SprintyCombat(CombatHandler):
    def __init__(self, client: wizwalker.client.Client, config_provider: CombatConfigProvider):
        super().__init__(client)
        self.config = config_provider

    async def get_card_named(self, name: str) -> Optional[CombatCard]:  # override to get rid of exception
        try:
            return await super().get_card_named(name)
        except ValueError:
            return None

    async def get_castable_cards(self) -> List[CombatCard]:  # extension for castable cards only
        cards = await self.get_cards()
        res = []
        for card in cards:
            if await card.is_castable():
                res.append(card)
        return res

    async def get_all_castable_cards_named(self, name: str) -> List[CombatCard]:
        cards = await self.get_castable_cards()
        res = []

        for card in cards:
            if name.lower() == (await card.name()).lower():
                res.append(card)

        return res

    async def get_castable_card_named(self, name: str) -> Optional[CombatCard]:  # extension to get only castable card
        cards = await self.get_castable_cards()

        for card in cards:
            if name.lower() == (await card.name()).lower():
                return card

        return None

    async def get_boss_or_none(self) -> Optional[CombatMember]:
        for m in await self.get_members():
            if m.is_boss():
                return m
        return None

    async def get_allies(self) -> List[CombatMember]:
        members = []
        my_client = await self.get_client_member()
        my_participant = await my_client.get_participant()
        my_team_id = await my_participant.team_id()
        my_id = await my_participant.owner_id_full()
        for mem in await self.get_members():
            participant = await mem.get_participant()
            if await participant.team_id() == my_team_id and await participant.owner_id_full() != my_id:
                members.append(mem)
        return members

    async def get_enemies(self) -> List[CombatMember]:
        members = []
        my_client = await self.get_client_member()
        my_participant = await my_client.get_participant()
        my_team_id = await my_participant.team_id()
        for mem in await self.get_members():
            participant = await mem.get_participant()
            if await participant.team_id() != my_team_id:
                members.append(mem)
        return members

    async def get_nth_ally_or_none(self, n: int) -> Optional[CombatMember]:
        allies = await self.get_allies()
        if len(allies) <= n:
            return None
        return allies[n]

    async def get_nth_enemy_or_none(self, n: int) -> Optional[CombatMember]:
        enemies = await self.get_enemies()
        if len(enemies) <= n:
            return None
        return enemies[n]

    async def try_execute_config(self, move_config: MoveConfig) -> bool:
        if move_config.move == "pass":  # switch to enum with real parser / better config format
            await self.pass_button()
            return True

        target: Any
        spell = move_config.move

        if move_config.target[0] is TargetType.Boss:
            boss = self.get_boss_or_none()
            if boss is None:
                return False
            target = boss
        elif move_config.target[0] is TargetType.Self:
            target = await self.get_client_member()
        elif move_config.target[0] is TargetType.AOE:
            target = None
        elif move_config.target[0] is TargetType.Enemy:
            if move_config.target[1] is None:
                target = self.get_nth_enemy_or_none(move_config.target[1])
            else:
                target = self.get_nth_enemy_or_none(0)
        elif move_config.target[0] is TargetType.Ally:
            if move_config.target[1] is not None:
                target = self.get_nth_ally_or_none(move_config.target[1])
            else:
                target = self.get_nth_ally_or_none(0)
        else:  # Forgot to implement or parsing failed
            raise NotImplementedError(str(move_config.target))

        if move_config.enchant != "none":
            enchant_card = await self.get_castable_card_named(move_config.enchant)
            if enchant_card is None:
                return False
            to_enchant = await self.get_castable_card_named(spell)
            if to_enchant is None:
                return False
            await enchant_card.cast(to_enchant)
            spellname = spell
            spell = None
            for s in await self.get_all_castable_cards_named(spellname):
                if await s.is_enchanted():
                    spell = s
                    break
        else:
            spell = await self.get_castable_card_named(spell)

        if spell is None:
            return False

        await spell.cast(target)
        return True

    async def handle_round(self):
        current_round = ((await self.round_number()) - 1) % self.config.defined_rounds()  # maybe account for fizzle?

        if round_config := self.config.get_round_or_none(current_round):
            for p in round_config:  # go through rounds priorities
                if await self.try_execute_config(p):
                    break  # we found a working priority and managed to cast it
            else:
                print(f"Ran out of priority moves for round {current_round}. Passing...")
                await self.pass_button()
        else:  # Very bad. Probably using empty config
            print(f"Full config fail! Passing...")
            await self.pass_button()
